"""
Utility class for storing information on calibration panels and combining measured radiance values with known
reflectance spectra.
"""

import matplotlib.pyplot as plt
from matplotlib import path
import numpy as np
from scipy.optimize import least_squares
import matplotlib.patches as patches
import cv2

import hylite
from hylite.reference.features import HyFeature
from hylite import HyData
from hylite.project import pix_to_ray_pano, pix_to_ray_persp


class Panel( HyData ):
    """
    A class for identifying calibration reference in images and storing
    the observed pixel radiances and known target reflectance values. This is used
    by, e.g., empirical line calibration procedures.
    """

    def __init__(self, material, radiance, strict=True, **kwds):
        """
        Generic constructor. Can be of the following forms:

        Args:
            material: a hylite.reference.Target instance containing target reflectance data for this panel.
            radiance: either a HyImage object (which contains some reference pixels) or a
                        NxM numpy array containing radiance values for N pixels across M bands.
            strict: True if measured radiance wavelengths must be entirely within the range of the reference material
                    spectra. Default is True. Use with care!
            **kwds: Keywords can include the following:

                 - wavelengths = wavelengths corresponding to the radiance values (if radiance is an array rather than a HyImage
                                 object).
                 - method = 'manual' (default) to manually pick reference in image (using an interactive plot). Can also be
                             'sobel' or 'laplace' to use the corresponding edge detectors to automatically identify the
                             calibration target (using OpenCV contouring).
                 - bands = list of band indices (integer) or wavelengths (float) to use when selecting the target.
                 - edge_thresh = the edge threshold (as percentile of gradient values) used when automatically identifying reference.
                                 Default is 95.
                 - area_thresh = the minimum area (in pixels) of candidate panels. Used to discard small/bright areas. Default is 100.
                 - shrink = a shrink factor to reduce the size of reference identified automatically (and so remove dodgy pixels
                            near the target edge/frame. Default is 0.4.
                 - db = If True, visualisation of the edge detection layers will be plotted for debug purposes. Default is false.
        """

        super().__init__(None)  # initialise header etc.
        self.source_image = None  # init defaults
        self.outline = None
        self.normal = None
        self.alpha = None
        self.skyview = None

        if isinstance(radiance, np.ndarray):  # radiance is a list of pixels

            # check and copy radiance data
            if len(radiance.shape) == 1:  # we've been given mean radiances
                radiance = radiance[np.newaxis, :]
            assert len(radiance.shape) == 2, "Error, radiance must be a 2-D array (N pixels by M bands)."
            self.data = radiance.copy()

            # check and copy wavelength data
            assert 'wavelengths' in kwds, "Error - wavelengths must be provided for pixel array."
            self.set_wavelengths(np.array(kwds["wavelengths"]))

        elif radiance.is_image():  # radiance is a hyimage

            self.source_image = radiance  # store reference to original image

            method = kwds.get("method", 'manual')  # what method to use to pick target?
            bands = kwds.get("bands", 428.0)

            # select target region
            if 'manual' in method.lower():  # pick region using interactive plot
                verts = radiance.pickPolygons(region_names=['Target'], bands=bands)[0]
                verts = np.vstack([verts, verts[0][None, :]])  # add return to first point
                self.outline = path.Path(verts)  # make matplotlib path from selected region

            else:
                db = kwds.get('db', False)  # draw plots?

                # calculate greyscale image
                if isinstance(bands, tuple) or isinstance(bands, list):
                    bands = [radiance.get_band_index(b) for b in bands]  # convert to indices
                    gray = np.sum(radiance.data[:, :, bands], axis=2) / np.nanmax(radiance.data[:, :, bands])
                else:
                    bands = radiance.get_band_index(bands)
                    gray = radiance.data[:, :, bands] / np.nanmax(radiance.data[:, :, bands])
                gray = cv2.GaussianBlur(gray, (3, 3), 0)  # apply slight blur to improve edge detection

                if db:
                    plt.figure(figsize=(20, 10))
                    plt.imshow(gray.T, cmap='gray')
                    plt.title("Greyscale")
                    plt.show()

                # extract edges
                if 'sobel' in method.lower() or 'auto' in method.lower():  # pick edges using sobel filter
                    sobelx = cv2.Sobel(gray.astype(np.float32), cv2.CV_64F, 1, 0, ksize=5)  # x
                    sobely = cv2.Sobel(gray.astype(np.float32), cv2.CV_64F, 0, 1, ksize=5)  # y
                    sobel = np.sqrt(sobelx ** 2 + sobely ** 2)
                    thresh = np.nanpercentile(sobel, kwds.get('edge_thresh', 95))
                    edge = sobel > thresh
                elif 'laplace' in method.lower():  # pick edges using laplace filter
                    laplacian = cv2.Laplacian(gray.astype(np.float32), cv2.CV_32F)
                    thresh = np.nanpercentile(laplacian, kwds.get('edge_thresh', 95))
                    edge = laplacian > thresh
                else:
                    assert False, "Error - %s is not a recognised extraction method. Try 'sobel' or 'laplace'." % method

                if db:
                    plt.figure(figsize=(20, 10))
                    plt.imshow(edge.T)
                    plt.title("Edge")
                    plt.show()

                # contour and find object contours
                _, threshold = cv2.threshold(edge.astype(np.uint8) * 254, 240, 255, cv2.THRESH_BINARY)
                _, contours, _ = cv2.findContours(threshold, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)

                # find brightest quadrilateral with area above threshold
                maxCtr = None
                maxBright = -1
                area_thresh = kwds.get("area_thresh", 100)
                for cnt in contours:

                    # simplify/approximate contour
                    approx = cv2.approxPolyDP(cnt, 0.1 * cv2.arcLength(cnt, True), True)

                    # is it a quadrilateral?
                    if approx.shape[0] == 4:
                        # calculate area
                        area = cv2.contourArea(approx)
                        if area > area_thresh:  # large enough?
                            # calculate approx brightness (by summing bounding box)
                            verts = np.array([approx[:, 0, 1], approx[:, 0, 0]]).T
                            xmin, xmax = np.min(verts[..., 0]), np.max(verts[..., 0])
                            ymin, ymax = np.min(verts[..., 1]), np.max(verts[..., 1])
                            patch = gray[xmin:xmax, ymin:ymax]
                            bright = np.nanmedian(patch)

                            if db:
                                plt.imshow(gray.T, cmap='gray')
                                plt.title("Candidate panel")
                                plt.axvline(xmin, color='r')
                                plt.axvline(xmax, color='r')
                                plt.axhline(ymin, color='r')
                                plt.axhline(ymax, color='r')
                                plt.show()

                            # store?
                            if maxBright < bright:
                                maxCtr = approx
                                maxBright = bright

                # convert to maplotlib path
                verts = np.array([maxCtr[:, 0, 1], maxCtr[:, 0, 0]]).T
                verts = np.vstack([verts, verts[0][None, :]])  # add return to first point
                self.outline = path.Path(verts, closed=True)  # make matplotlib path from selected region

                # shrink to 40% of original size (to remove frame and dodgy edge pixels)
                centroid = np.mean(self.outline.vertices[1::, :], axis=0)
                verts = kwds.get("shrink", 0.4) * (self.outline.vertices - centroid) + centroid
                self.outline = path.Path(verts, closed=True)

            # calculate pixels within selected region
            xx, yy = np.meshgrid(np.arange(radiance.xdim()), np.arange(radiance.ydim()))
            xx = xx.flatten()
            yy = yy.flatten()
            points = np.vstack([xx, yy]).T  # coordinates of each pixel
            mask = self.outline.contains_points(points)  # identify points within path
            mask = mask.reshape((radiance.ydim(), radiance.xdim())).T  # reshape to pixel mask

            # extract pixel reflectance values based on this mask
            self.data = np.array([radiance.data[:, :, b][mask] for b in range(radiance.band_count())]).T

            # also copy across wavelength info
            assert radiance.has_wavelengths(), "Error - radiance image must have wavelength information."
            self.set_wavelengths(radiance.get_wavelengths())  # get wavelength data

        else:
            assert False, "Error: radiance argument must be a HyImage instance or a numpy array of pixels."

        # if we have lots of target pixels (we don't need that many), only keep top 50% [as darker ones likely result from
        # dodgy border effects
        if self.data.shape[0] > 30:
            brightness = np.nanmean(self.data, axis=1)
            mask = brightness > np.nanpercentile(brightness, 50)
            self.data = self.data[mask, :]

        # extract reflectance data from target
        target_bands = material.get_wavelengths()
        if strict:
            assert np.nanmin(target_bands) <= np.nanmin(
                self.get_wavelengths()), "Error - calibration range does not cover pixel range. " \
                                         "Radiance data starts at %.1f nm but calibration data starts %.1f nm." % (
                                             np.nanmin(self.get_wavelengths()), np.nanmin(target_bands))
            assert np.nanmax(target_bands) >= np.nanmax(
                self.get_wavelengths()), "Error - calibration range does not cover pixel range. " \
                                         "Radiance data ends at %.1f nm but calibration data ends %.1f nm." % (
                                             np.nanmax(self.get_wavelengths()), np.nanmax(target_bands))
        idx = [np.argmin(np.abs(target_bands - w)) for w in self.get_wavelengths()]  # matching wavelengths
        self.reflectance = material.get_reflectance()[idx]
        self.material = material

    def set_outline(self, coords ):
        """
        Define a panel outline (for e.g., estimating orientation) based on a known camera pose.

        Args:
            coords: the coordinates of the four panel corners.
        """
        assert len(coords) == 4 and len(coords[0]) == 2, "Error - coords must have shape (4,2) not %s" % str( np.array(coords).shape )
        self.outline = path.Path(coords, closed=True)

    def copy(self):
        """
        Make a deep copy of this panel instance.

        Returns:
             a new Panel instance.
        """
        return Panel( self.material, self.data, wavelengths=self.get_wavelengths() )

    def get_mean_radiance(self):
        """
        Calculate and return the mean radiance for each band of all the pixels in this calibration region.
        """
        return np.nanmean(self.data, axis=0)

    def get_reflectance(self):
        """
        Get the known (reference) reflectance of this panel.
        """
        return self.reflectance

    def get_normal(self, cam=None, recalc=False):
        """
        Get the normal vector of this panel by assuming its outline is square (prior to projection onto the camera).

        Args:
            cam: a Camera object describing the pose of the camera from which the panel is viewed. Default is None (to retrieve
                 previously stored normals)
            recalc: True if a precomputed (or otherwise defined) normal vector should be recalculated. Default is False.
        Returns:
            the normal vector of the panel (in world coordinates). This is also stored as self.normal.
        """

        if recalc or self.normal is None:

            # check outline is available
            assert cam is not None, "Error - normal vector not previously defined. Please specify a camera object (cam) to estimate normal."
            assert self.outline is not None, "Error - self.outline must contain four points to estimate this panels normal vector..."

            # get corners of panel and convert to rays
            corners = np.array([self.outline.vertices[i, :] for i in range(4)])

            if cam.is_panoramic():
                ray1 = pix_to_ray_pano(corners[0, 0], corners[0, 1], cam.fov, cam.step, cam.dims)
                ray2 = pix_to_ray_pano(corners[1, 0], corners[1, 1], cam.fov, cam.step, cam.dims)
                ray3 = pix_to_ray_pano(corners[2, 0], corners[2, 1], cam.fov, cam.step, cam.dims)
                ray4 = pix_to_ray_pano(corners[3, 0], corners[3, 1], cam.fov, cam.step, cam.dims)
            else:
                ray1 = pix_to_ray_persp(corners[0, 0], corners[0, 1], cam.fov, cam.dims)
                ray2 = pix_to_ray_persp(corners[1, 0], corners[1, 1], cam.fov, cam.dims)
                ray3 = pix_to_ray_persp(corners[2, 0], corners[2, 1], cam.fov, cam.dims)
                ray4 = pix_to_ray_persp(corners[3, 0], corners[3, 1], cam.fov, cam.dims)

            a = 1.0  # length of each square (in arbitrary coordinates)
            h = np.sqrt(2)  # length of hypot relative to sides

            def opt(x, sol=False):
                # get test depths
                z1, z2, z3, z4 = x

                # calculate edge coordinates
                A = ray1 * z1
                B = ray2 * z2
                C = ray3 * z3
                D = ray4 * z4

                # and errors with edge lengths
                AB = np.linalg.norm(B - A)
                BC = np.linalg.norm(C - B)
                CD = np.linalg.norm(D - C)
                DA = np.linalg.norm(A - D)
                AC = np.linalg.norm(C - A)
                BD = np.linalg.norm(D - B)

                if not sol:
                    return [AB - a, BC - a, CD - a, DA - a, AC - h, BD - h]  # return for optimiser
                else:  # return solution (normal vector)
                    AB = (B - A) / AB
                    BC = (C - B) / BC
                    return np.cross(AB, BC)

            # get normal vector in camera coords
            sol = least_squares(opt, (10.0, 10.0, 10.0, 10.0))
            norm = opt(sol.x, sol=True)

            # rotate to world coords
            norm = np.dot(cam.get_rotation_matrix(), norm)
            self.set_normal(norm)

        return self.normal

    def set_normal(self, n ):
        """
        Set panel normal vector to a known vector.

        Args:
            n: a (3,) numpy array containing the normal vector in world coordinates.
        """
        if n is None:
            self.normal = None # remove normal
        else:
            assert len(n) == 3, "Error - n must be a (3,) normal vector."
            self.normal = np.array(n) / np.linalg.norm(n) # enforce n has length 1.0
            if self.normal[2] < 0:
                self.normal *= -1 # panel always points upwards

    def get_skyview(self, hori_elev=0.0, up=np.array([0, 0, 1])):
        """
        Get this panels skyview factor. Normal vector must be defined, otherwise an error will be thrown.

        Args:
            hori_elev: the angle from the panel to the horizon (perpendicular to the panel's orientation) in degrees.
                       Used to reduce the sky view factor if the panel is below the horizon (e.g. in an open pit mine).
                       Default is 0.0 (i.e. assume a flat horizon). Can also be negative if sky is visible below the
                       (flat) horizon.
            up: the vertical (up) vector. Default is [0,0,1].
        Returns:
            this panels sky view factor (assuming the panel is relatively unoccluded and the horizon is flat).
        """

        # proportion of sky visible assuming horizontal horizon
        s = (np.pi - np.arccos(np.dot(up, self.normal))) / np.pi

        # adjust according to hori_elev [ and enforce range from 0 - 1.0
        self.skyview = min(max(0, s - (np.deg2rad(hori_elev) / np.pi)), 1.0)
        return self.skyview

    def get_alpha(self, illudir):
        """
        Return the reflected light fraction of this panel based on the specified illumination direction using
        Lamberts' cosine law.
        """
        assert len(illudir) == 3, "Error - illudir must be a (3,) numpy array."
        assert self.normal is not None, 'Error - normal vector must be defined to estimate alpha.'
        if illudir[2] > 0: # check illudir is pointing downwards
            illudir = illudir * -1
        self.alpha = max( 0, np.dot( -self.normal, illudir ) )
        return self.alpha

    def quick_plot(self, bands=hylite.RGB, **kwds):

        """
        Quickly plot the outline of this calibration target for quality checking etc.

        Args:
            bands: the image bands to plot as a preview. Default is io.HyImage.RGB.
            **kwds: keywords are passed to HyData.plot_spectra( ... ).
        """
        if self.source_image is not None: # plot including preview of panel
            fig, ax = plt.subplots(1, 2, figsize=(15, 5), gridspec_kw={'width_ratios': [1, 3]})

            # plot base image
            self.source_image.quick_plot(bands, ax=ax[0])
            ax[0].set_xticks([])
            ax[0].set_yticks([])

            # plot target on image and set extents
            patch = patches.PathPatch(self.outline, edgecolor='orange', fill=False, lw=2)
            bbox = self.outline.get_extents()
            padx = (bbox.max[0] - bbox.min[0]) * 1.5
            pady = (bbox.max[1] - bbox.min[1]) * 1.5
            ax[0].add_patch(patch)
            ax[0].set_xlim(bbox.min[0] - padx, bbox.max[0] + padx)
            ax[0].set_ylim(bbox.max[1] + pady, bbox.min[1] - pady)

            # plot spectra
            kwds['labels'] = kwds.get('labels', HyFeature.Themes.ATMOSPHERE)
            self.plot_spectra( ax=ax[1], **kwds )
        else: # no image data, just plot spectra
            kwds['labels'] = kwds.get('labels', HyFeature.Themes.ATMOSPHERE)
            fig, ax = self.plot_spectra(**kwds)
            ax.set_ylabel('Downwelling Radiance')
        return fig, ax

    def plot_ratio(self, ax = None):

        """
        Plots the ratio between known reflectance and observed radiance for each band in this target.

        Args:
            ax: the axes to plot on. If None (default) then a new axes is created.
        Returns:
            fig, ax = the figure and axes objects containing the plot.
        """

        if ax is None:
            fig, ax = plt.subplots(figsize=(15, 10))

        # calculate ratio
        ratio = self.get_mean_radiance() / self.get_reflectance()

        # plot
        ax.plot( self.get_wavelengths(), ratio )
        ax.set_ylabel("radiance / reflectance" )
        ax.set_xlabel("Wavelength (nm)")

        return fig, ax




"""
Read ENVI header files.
"""

import os
import glob
import numpy as np
from hylite import HyHeader
import re

def makeDirs(path):
    """
    Checks if all the directories to the specified file exist, and if they do not, makes them.

    Args:
        path: the directory or file path to check. If any directories do not exist, they will be created.
    """

    # convert path to directory
    if not os.path.isdir(path):
        path = os.path.dirname(path)

    # make directory if need be
    if not os.path.exists(path) and path != '':
        os.makedirs(path)

def saveHeader(path, header):
    """
    Write to a header file.

    Args:
        path: the path to write to.
        header: the HyHeader object to write.
    """

    if os.path.exists(path):
        os.remove(path)  # delete existing header so we can overwrite it (causes weird bug if we don't....)

    with open( os.open( path, os.O_CREAT | os.O_WRONLY, 0o777) , "w+") as f: # create file with appropriate permissions
        f.write("ENVI\n")
        for key, value in header.items():

            #get value as string
            if isinstance(value, list) or isinstance(value, np.ndarray): #convert lists to string
                if isinstance(value[0], str):  # format a list of strings
                    s = ', '.join(value)
                else:  # format a list of numbers
                    s = ("%s" % list(value))[1:-1]  # convert to string and strip '[' and ']' characters
            else:
                s = str(value).strip()

            #write in the relevant format
            if s == "":
                f.write("%s = { }\n" % key)
            elif "," in s:
                if len(s) < 512:
                    f.write("%s = {%s}\n" % (key, s))
                else:
                    f.write("%s = {" % key)
                    splt = s.split(",")
                    for i, l in enumerate(splt[:-1]):
                        if i % 10 == 0: # new line every 10th value
                            f.write("\n")
                        f.write("%s," % l) # write values
                    f.write("%s}\n" % splt[-1])
            else:
                f.write("%s = %s\n" % (key, s))
        f.close()

def loadHeader(path):
    """
    Load a header file.

    Args:
       file: a file path to a .hdr file.
    """
    header = HyHeader()

    # store path
    header['path'] = path

    inblock = False
    try:
        hdrfile = open(path, "r")
    except:
        if path is not None: # no header file exists. This is not necessarily an issue.
            print("Could not open hdr file %s" % str(path))
        return None

    # Read line, split it on equals, strip whitespace from resulting strings and add key/value pair to output
    currentline = hdrfile.readline()
    while (currentline != ""):
        # ENVI headers accept blocks bracketed by curly braces - check for these
        if not inblock:
            # Split line on first equals sign
            if (re.search("=", currentline) is not None):
                linesplit = re.split("=", currentline, 1)
                # key = str.lower(linesplit[0].strip())
                key = linesplit[0].strip()
                value = linesplit[1].strip()

                # If value starts with an open brace, it's the start of a block - strip the brace off and read the rest of the block
                if (re.match("{", value) is not None):
                    inblock = True
                    value = re.sub("^{", "", value, 1)
                    # If value ends with a close brace it's the end of the block as well - strip the brace off
                    if (re.search("}$", value)):
                        inblock = False
                        value = re.sub("}$", "", value, 1)
                value = value.strip()
                header[key] = value
        else:

            # If we're in a block, just read the line, strip whitespace (and any closing brace ending the block) and add the whole thing
            value = currentline.strip()
            if (re.search("}$", value)):
                inblock = False
                value = re.sub("}$", "", value, 1)
                value = value.strip()
            header[key] = header[key] + value
        currentline = hdrfile.readline()
    hdrfile.close()

    #convert default things like wavelength data to numeric form
    #N.B. wavelength should ALWAYS be stored as nanometres
    if 'Wavelength' in header: # drop upper case wavelength for some files
        header['wavelength'] = header['Wavelength']
        del header['Wavelength']
    if "wavelength" in header:
        units = header.get("wavelength units", "nm").lower()
        if "nm" in units or "nano" in units: #units in nanometers
            header['wavelength'] = np.fromstring(header['wavelength'], sep=',')
        elif "um" in units or "micro" in units:
            header['wavelength'] = np.fromstring(header['wavelength'], sep=',') * 1000.0
        elif "mm" in units or "milli" in units:
            header['wavelength'] = np.fromstring(header['wavelength'], sep=',') * 1000000.0
        elif "wavenum" in units or "cm-1" in units:
            header['wavelength'] = (1 / np.fromstring(header['wavelength'], sep=',')) * 10000000.0
        elif "unk" in units.lower():
            print("Warning - unknown wavelength units. Assuming nanometers.")
            header['wavelength'] = np.fromstring(header['wavelength'], sep=',')
        else:
            assert False, "Error - unrecognised wavelength format %s." % units

        header['wavelength units'] = 'nm' #update wavelength units

    #if "band names" in header:
    #    header["band names"] = header['band names'].split(',') #split into list

    if "fwhm" in header:
        header["fwhm"] = np.fromstring(header['fwhm'], sep=',') #split into list

    if "bbl" in header:
        header["bbl"] = np.fromstring(header['bbl'], sep=',') #split into list

    return header

def matchHeader(path):
    """
    Matches image and header path.

    Args:
        path: the path to an image or header file

    Returns:
        A tuple containing:

        - header = file path to the associated .hdr or .HDR file (if found, otherwise None)
        - image = file path to the associated image data (if found, otherwise None).
    """

    # find files with the same name but different extensions
    path, ext = os.path.splitext(path)
    header = None
    image = None
    match = glob.glob(path + "*")

    assert (path + ext) in match, "Error - file not found (%s)" % (path + ext)
    match.remove(path + ext)  # remove self-match

    # we have a header file, find associated image
    if "hdr" in str.lower(ext):
        header = path + ext  # store header file

        # did we find image data?
        for m in match:
            # ignore potentially associated file types (that aren't the data file we're looking for)
            if os.path.splitext(m)[0] == path \
                    and not "log" in str.lower(os.path.splitext(m)[1]) \
                    and not "png" in str.lower(os.path.splitext(m)[1]) \
                    and not "jpg" in str.lower(os.path.splitext(m)[1]) \
                    and not "bmp" in str.lower(os.path.splitext(m)[1]) \
                    and not "hdt" in str.lower(os.path.splitext(m)[1]) \
                    and not "csv" in str.lower(os.path.splitext(m)[1]) \
                    and not "txt" in str.lower(os.path.splitext(m)[1]) \
                    and not "xml" in str.lower(os.path.splitext(m)[1]) \
                    and not "cam" in str.lower(os.path.splitext(m)[1]) \
                    and not "brm" in str.lower(os.path.splitext(m)[1]):

                image = m  # store matching image file
                break
            elif (os.path.splitext(m)[0] == path) and (image is None):
                image = m # .png files can be valid, but should be lower priority, so store but continue looking
                          # in case there is e.g., a .dat file also.
    # we have an image file, find associated header file
    else:
        image = path + ext
        for m in match:
            if ".hdr" in str.lower(m) and os.path.splitext(m)[0] == path:
                header = m
                break

    return header, image




"""
Create, store and manipulate spectral libraries.
"""

from hylite.hydata import HyData
from hylite.hyfeature import HyFeature, MultiFeature, MixedFeature
import hylite.reference.features as ref
import numpy as np
import matplotlib.pyplot as plt

def from_indices(data, indices, s=4, names=None, ):
    """
    Extract a spectral library by sampling and averaging pixels within the specified distance of sample points.
    Args:
        data (HyData): a HyData instance containing the spectral data.
        indices (list): a list of sample indices to extract spectra from
        s (int): the number of adjacent points to include in each sample. For HyImage data this will be a square patch of sxs
           pixels. For HyCloud data s is used to take adjacent points from the points list (which is assumed to be somewhat ordered).
        names (list): a list containing names for each sample, or None to generate (numeric) sample ids.

    Returns:
        a HyLibrary instance
    """

    if names is None:
        names = ["Sample_%d" % i for i in np.arange(1, len(indices) + 1)]

    # extract spectra
    S = []
    for idx, name in zip(indices, names):
        if isinstance(idx, int):
            X = data.data[idx - s: idx + 1, :]
        elif len(idx) == 2:  # image
            X = data.data[idx[0] - s: idx[0] + s, idx[1] - s: idx[1] + s, :].reshape(-1, data.band_count())
        elif len(idx) == 1:
            X = data.data[idx[0] - s: idx[0] + s, :]
        S.append(HyLibrary(X[None, :, :], lab=[name], wav=data.get_wavelengths()))

    # merge
    out = S[0]
    for i in range(1, len(S)):
        out += S[i]

    # return
    return out

def from_classification(data, labels, names=None, ignore=[0], subsample='all'):
    """
    Extract a spectral library from a labelled dataset.

    Args:
        data (HyData): a HyData instance containing the hyperspectral data.
        labels (HyData): a HyData instance or numpy array containing class labels.
        names (dict): a dictionary with numerical labels (integers) in labels.data as keys and corresponding string names
               as values. If None, class names will be pulled from the header, and if these don't exist then integer
               names will be used.
        ignore (list): a list of class IDs (integer values) to ignore. By default all nans and 0's will be ignored.
        method (str): method used for reducing the number of spectra in this library. Options are 'all' (retain all
                measurements), an integer of the number of points to (randomly) sample, or a list containing
                percentiles to sample. Using the percentile method, samples will be ranked by median brightness
                (albedo) and the spectra corresponding to the desired percentiles kept in the library. Default is 'all'.

    Returns:
        a HyLibrary instance.
    """

    X = data.X()

    if isinstance(labels, HyData):
        L = labels.X().ravel()
    else:
        assert isinstance(labels, np.ndarray), "Error - labels must be array or HyData, not %s" % str(type(labels))
        L = labels.ravel()
    assert X.shape[0] == L.shape[0], "Error - number of spectra (%d) does not match number of labels (%d)?" % (
    X.shape[0], L.shape[0])

    ids = np.unique(L)  # get unique classes

    # build name dictionary if not provided
    cls_names = {}
    if names is None:
        classes = np.arange(0, np.nanmax(ids))
        if isinstance(labels, HyData) and 'class names' in labels.header:
            classes = labels.header['class names']
            if isinstance(classes, str):
                classes = classes.split(',')
        for i, n in enumerate(classes):
            cls_names[i] = n

    Y = []  # store columns of spectral library here
    samples = []  # sample names
    for i, n in enumerate(ids):  # loop through classes
        if n in ignore:
            continue  # skip ignored classes

        _S = X[L == n, :]  # get samples
        if isinstance(subsample, str) and 'all' in subsample:
            S = _S  # keep all samples
        if isinstance(subsample, int):  # random selection
            if _S.shape[0] < subsample:
                S = _S
            else:
                S = _S[np.random.choice(subsample, subsample), :]  # select random spectra
        elif isinstance(subsample, list) or isinstance(subsample, tuple) or isinstance(subsample, np.ndarray):
            # get percentiles
            a = np.nanmean(_S, axis=-1)
            idx = np.argsort(a)
            S = []
            for p in subsample:
                assert isinstance(p, int), "Error - percentiles must be integers between 0 and 100, not %s." % p
                assert 0 <= p and 100 >= p, "Error - percentiles must be integers between 0 and 100, not %s." % p
                p = int(p / 100. * len(idx))  # convert percentile to index
                S.append(_S[idx[p], :])
            S = np.array(S)
        # store
        Y.append(S)

        samples.append(cls_names.get(n, 'S_%s' % (i + 1)))

    out = np.full((len(Y), np.max([len(y) for y in Y]), X.shape[-1]), np.nan)  # build output array
    for i, s in enumerate(Y):
        out[i, :s.shape[0], :] = s

    return HyLibrary(out, samples, wav=data.get_wavelengths())

class HyLibrary(HyData):
    """
    A class for loading and managing spectral libraries and associated metadata.
    """

    def __init__(self, data, lab=None, wav=None, header=None):
        """
        Args:
            data (ndarray): a numpy array containing spectral measurements and indexed as [sample, measurement, band]. If different
                  numbers of measurements are available per sample then some of the measurement axes can be set as nan. If
                  a 2D array is passed (sample, band) then it will be expanded to (sample, 1, band).
            lab (list): list of sample labels (one label per sample). Default is None (labels will be integers from 0 - n).
            wav (list, ndarray): list of wavelengths in the spectra for each sample. If None this must be defined in the header file passed.
            header (hylite.HyHeader): a HyHeader instance containing additional metadata to associate with this library.
        """

        # checks
        assert isinstance(data, np.ndarray), "Error - reflectance data must be a numpy array."
        if len(data.shape) == 2:
            data = data[:,None,:] # expand to correct shape
        assert len(data.shape) == 3, "Error - reflectance data must by a 2D array indexed as [sample,measurement,bands]."

        # init HyData object with reflectance data
        super().__init__(data, header=header)
        self.header['file type'] = 'Hylite Library'  # set file type

        # store wavelength data
        if wav is None:
            assert self.header.has_wavelengths(), "Error - no wavelength data specified"
        else:
            assert isinstance(wav, list) or isinstance(wav, np.ndarray), "Error - wavelengths must be a list."
            assert data.shape[-1] == len(wav), "Error - wavelengths must be specified for each reflectance band."
            self.header['wavelength'] = wav.copy()  # store wavelength data

        # store sample labels
        if lab is not None:
            self.set_sample_names( lab )

    def copy(self,data=True):
        """
        Make a deep copy of this HyLibrary instance.

        Args:
            data (bool): True if a copy of the data should be made, otherwise only copy header.

        Returns:
            a new HyLibrary instance.
        """
        header = self.header.copy()
        if data:
            arr = self.data.copy()
        else:
            arr = self.data
        names = None
        wav = None
        if self.has_sample_names():
            names = self.get_sample_names()
        if data==True and self.has_wavelengths():
            wav = self.get_wavelengths()
        out = HyLibrary(arr, lab=names, wav=wav, header=header)
        if not data:
            out.data = None # drop data array to protect it (we don't want shallow copies)
        return out

    def as_image(self, shallow=False):
        """
        Convert this library to a HyImage dataset for e.g. visualisation using HyImage.quick_plot(...).

        Args:
            shallow (bool): True if the underlying data array should be shared. Default is False as copies are dangerous.

        Returns:
            a HyImage instance of this dataset.
        """
        from hylite.hyimage import HyImage # N.B. this import must be here to avoid circular import
        if shallow:
            return HyImage( self.data, header=self.header )
        else:
            return HyImage( self.data.copy(), header=self.header.copy() )

    def merge(self, library2):
        """
        Return a copy of this library merged with another.
        """

        assert (self.get_wavelengths() == library2.get_wavelengths()).all(), "Error - libraries must have same wavelength info..."

        # copy/merge data
        out = self.copy()
        out.data = np.vstack( [self.data, library2.data] )
        if (self.upper is not None) and (self.lower is not None):
            out.upper = np.vstack( [self.upper, library2.upper] )
        if (self.lower is not None) and (self.lower is not None):
            out.lower = np.vstack([self.lower, library2.lower])
        if self.has_sample_names() and library2.has_sample_names():
            out.set_sample_names ( list(self.get_sample_names())+list(library2.get_sample_names()))

        return out

    def sample_count(self):
        """
        Number of samples/spectra in this library.
        """
        return self.data.shape[0]

    def has_sample_names(self):
        """
        Has this library got associated sample names?
        """
        return 'sample names' in self.header

    def get_sample_names(self):
        """
        Get list of sample names.
        """
        if self.has_sample_names():
            return self.header.get_list('sample names') # names are defined
        else:
            # no names defined, so make some
            self.header['sample names'] = ['S%d' % i for i in range(self.sample_count()) ]
            return self.get_sample_names()

    def set_sample_names(self, names):
        """
        Set sample names
        """
        assert isinstance(names, list) or isinstance(names, np.ndarray), "Error - sample labels must be a list."
        assert self.data.shape[0] == len(names), "Error - sample names must be specified for each sample."
        self.header['sample names'] = np.array(names)

    def has_groups(self):
        """
        Returns true if any groups are present in this dataset.
        """
        return len(self.get_groups()) > 0

    def get_groups(self):
        """
        Return a list of groups that have been defined for the HyLibrary.
        """
        groups = []
        for k,v in self.header.items():
            k = k.split(' ')
            if k[0].strip() == 'group':
                groups.append( k[1].strip() )
        return groups

    def get_group_ids(self, name ):
        """
        Return a list containing the indices of spectra associated with the specified group.  This can be useful for
        slicing the self.data matrix.

        Args:
            name (str): a string containing the name of the group.

        Returns:
            a list containing integer indices of the spectra in this group.
        """
        assert 'group %s'%name in self.header, "Error - no group named %s exists." % name
        return self.header.get_list('group %s'%name)

    def get_group(self, name, shallow=False):
        """
        Return a HyLibrary instance containing only the spectra associated with the specified group.

        Args:
            name (str): the name of the group.
            shallow (bool): True if the returned HyLibrary instance should be a shallow copy (share the same underlying data
                      array and header file) as this instance so changes propagate. Default is False - shallow copies
                      are powerful but dangerous!.
        Returns:
            a HyLibrary instance containing the requested group of spectra.
        """
        ids = self.get_group_ids(name)
        ids = [ self.get_sample_index(i) for i in ids ]
        if shallow:
            return HyLibrary( self.data[ids,:], lab=self.get_sample_names()[ids], wav=self.get_wavelengths() )
        else:
            return HyLibrary( self.data[ids,:].copy(),  lab=self.get_sample_names()[ids], wav=self.get_wavelengths() )

    def add_group(self, name, indices):
        """
        Associate the specified spectra with some arbitrary grouping key for easy subsequent access. Note that if
        a group of this name already exists then it will be overwritten / updated.

        Args:
            name (str): the name of this group.
            indices (list, ndarray): the indices (integer) or sample names (string) associated with each of the
                     spectra in this group.
        """
        name = name.strip()
        assert name not in self.get_sample_names(), "Error - %s is already a sample name." % name
        self.header['group %s'%name] = indices # add group to header file

    def __getitem__(self, n):
        """
        Slice this library to extract groups or label names. Keys can be either string defining sample or group names,
        integers (treated as indices of the sample array) or lists/arrays of the above.

        Args:
            n (str): the group or label name to extract.
        """
        if isinstance(n, list) or isinstance(n, tuple) or isinstance(n, np.ndarray):
            if len(n) == 0:
                assert False, 'Error - %s is an invalid spectra key.' % str(n)
            out = self[n[0]] # create ouput with first entry
            for i in range(1, len(n)):
                out += self[n[i]] # append each entry
            return out # done, return
        else:
            if isinstance(n, int) or isinstance(n,float): # subset using index
                n = self.get_sample_names()[int(n)] # get corresponding name
            if n in self.get_sample_names(): # this is a sample name
                idx = self.get_sample_index(n)
                arr = self.data[ [idx], :, : ]
            elif isinstance(n, str):
                return self.get_group( n ) # return group (easy)
            else:
                assert False, "Error - %s is an invalid spectra key." % str(n)
            return HyLibrary(arr.copy(), lab= [n], wav=self.get_wavelengths())

    def __add__(self, other):

        if (other.band_count() != self.band_count()):
            print("Warning - wavelength arrays do not match. Resampling.")
            other = other.resample( self.get_wavelengths(), vb=False )

        # merge samples
        S = {}
        maxs = -np.inf
        for n in self.get_sample_names():
            arr = self[n].data[0, ...]
            if n in S:
                S[n].append(arr)
            else:
                S[n] = [arr]
            maxs = max(maxs, np.sum([a.shape[0] for a in S[n]]))
        for n in other.get_sample_names():
            arr = other[n].data[0, ...]
            if n in S:
                S[n].append(arr)
            else:
                S[n] = [arr]
            maxs = max(maxs, np.sum([a.shape[0] for a in S[n]]))

        # stack into output array
        arr = np.full( (len(S), maxs, self.band_count() ), np.nan )
        names = []
        for i, (k, v) in enumerate(S.items()):
            names.append(k)
            sample = np.vstack( v )
            arr[i, :sample.shape[0], : ] = sample

        # return
        return HyLibrary( arr, lab=names, wav=self.get_wavelengths())

    # noinspection PyChainedComparisons
    def get_sample_index(self, name):
        """
        Get the index associated with a sample name.

        Args:
            name (str, int): the desired sample name. If an integer passed it is returned (assumed to be an index).

        Returns:
            the sample index (int).
        """
        if np.issubdtype( type(name), np.integer):  # already an int - easy!
            idx = name
            # noinspection PyChainedComparisons
            assert 0 <= idx < len(self.get_sample_names()), "Error - %d is an invalid index." % idx
            return idx
        else:
            assert name in self.get_sample_names(), "Error - invalid sample name %s" % name
            return list(self.get_sample_names()).index(name)

    def get_sample_spectra(self, sample ):
        """
        Get the reflectance spectra of the specified sample.

        Args:
            sample (str, int): the sample name or index to extract.

        Returns:
            a numpy array containing the specified spectra.
        """

        idx = self.get_sample_index(sample)
        return self.data[ idx, : ]

    def collapse(self):
        """
        Returns a copy of this HyLibrary with any groups collapsed into individual samples.
        """
        groups = self.get_groups()
        assert len(groups) > 0, "Error - library has no groups to collapse."

        # get each group and collapse into samples
        samples = []
        for g in groups:
            samples.append( np.vstack( self.get_group(g).data ) ) # stack group

        # build new data array
        arr = np.full( (len(groups), np.max([s.shape[0] for s in samples]), self.band_count()), np.nan)
        for i,v in enumerate(samples):
            arr[i, 0:v.shape[0], : ] = v

        # return output
        return HyLibrary( arr, lab=groups, wav=self.get_wavelengths() )

    def squash(self):
        """
        Returns a copy of this HyLibrary instance with multiple measurements averaged to give a single spectra per sample.
        """

        data = np.nanmedian(self.data, axis=1)
        out = self.copy(data=False)
        out.data = data[:, None, :]
        return out


    def quick_plot(self, band_range=None, ax=None, labels=None, pad=None, collapse=False, hc=False, **kwds):
        """
        Plots individual spectra in this library.

        Args:
            ax: an axis to plot to. If None (default), a new axis is created.
            band_range (tuple): the (min,max) band index (int) or wavelength (float) to plot.
            labels (list): list of HyFeature instances to plot on spectra. Default is reference.features.DIAGNOSTIC
            pad (float): the spacing to add between individual spectra. Default is 5% of the range of reflectance values.
            collapse (bool): True if groups should be plotted rather than individual samples. Default is False.
            hc (bool): True if the plotted spectra should be hull corrected first. Default is False.
            **kwds: Keywords are passed to plt.plot( ... ), except the following:

                 - clip = a tuple with the min, med and max percentiles to plot. Default is (5,50,95).
                 - figsize = a figsize for the figure to create (if ax is None).

        Returns:
            Tuple containing

            - fig: a matplotlib figure instance.
            - ax: a matplotlib axes.
        """

        if collapse:
            self = self.collapse() # this seems dodgy somehow?

        if labels is None:  # allow None labels (= don't plot labels)
            labels = []

        # parse band range
        minb = 0
        maxb = self.band_count() - 1
        if not band_range is None:
            minb = self.get_band_index(band_range[0])
            maxb = self.get_band_index(band_range[1])

        # hull correct?
        if hc:
            from hylite.correct import get_hull_corrected
            self = get_hull_corrected(self, band_range=(minb, maxb), vb=False)
            minb = 0 # spectra has already been subset
            maxb = self.band_count() - 1

        # create axes if need be
        if ax is None:
            height = max(min(self.data.shape[0], 100), 8)
            fig, ax = plt.subplots(figsize=kwds.pop('figsize', (18, height) ))

        # calculate pad if need be
        if pad is None:
            pad = 0.05 * (np.nanmax(self.data[..., minb:maxb]) - np.nanmin(self.data[..., minb:maxb]))

        # plot spectra
        wav = self.get_wavelengths()[minb:maxb]
        baseline = np.zeros(len(wav))
        yticks = []
        color_map = kwds.pop('color', ['g', 'b'])
        if not isinstance(color_map, list) or isinstance(color_map, np.ndarray):
            color_map = list(color_map)

        clip = kwds.pop('clip', (5,50,95))  # get percentiles keyword for plotting
        for i in range(self.data.shape[0]):

            # extract spectra slice
            refl = self.data[i, :, minb:maxb]
            y0 = y1 = None
            if refl.shape[0] > 1:  # plot percentiles
                if isinstance(clip, int):
                    y = np.nanpercentile(refl, clip, axis=0)
                elif len(clip) == 1:
                    y = np.nanpercentile(refl, clip[0], axis=0)
                elif len(clip) == 3:
                    y0, y, y1 = np.nanpercentile(refl, clip, axis=0)
                else:
                    assert False, "%s is an invalid clip. Must be either (min,med,max) or an integer." % str(clip)
            else:
                y = refl[0, :]


            # calculate offset
            if y0 is None:
                offset = np.abs(np.nanmax(np.abs(y - baseline))) + pad
            else:
                offset = np.abs(np.nanmax(np.abs(y0 - baseline))) + pad

            # calculate color
            kwds['color'] = color_map[i % len(color_map)]

            # plot spectra
            _y = offset + y  # calculate y
            ax.plot(wav, _y, **kwds)
            baseline = _y
            yticks.append(_y[0])

            # plot range
            if (y0 is not None) and (y1 is not None):
                _y = np.hstack([offset + y0, (offset + y1)[::-1]])
                _x = np.hstack([wav, wav[::-1]])
                ax.fill(_x, _y, color='grey', alpha=0.25)
                baseline = offset + y1  # move baseline up to upper error bound

        # ax.autoscale(False)
        i = 0
        if not isinstance(labels, list): labels = [labels]
        for l in labels:

            # plot label
            if ax.get_xlim()[0] < l.pos < ax.get_xlim()[1]:
                if isinstance(l, MultiFeature) or isinstance(l, MixedFeature):
                    l.quick_plot(ax=ax, method='fill+line', sublabel=i)
                    i += l.count()
                else:
                    l.quick_plot(ax=ax, method='fill+line', label=i)
                    i += 1

        # sort out y-ticks
        ax.set_yticks(yticks)
        ax.set_yticklabels(self.get_sample_names())

        # add major x-ticks
        mn, mx = ax.get_xlim()
        ticks = range(int(mn / 100) * 100, int(mx / 100) * 100, 100)
        ax.set_xticks(ticks)
        ax.set_xticklabels(ticks, rotation=45)

        # add minor x-ticks
        order = int(np.log10(ax.get_xlim()[1] - ax.get_xlim()[0])) - 1
        mn = np.around(ax.get_xlim()[0], -order)
        mx = np.around(ax.get_xlim()[1], -order)
        ax.set_xticks(np.arange(mn, mx, 10 ** order)[1::], minor=True)
        ax.set_xticklabels([], minor=True)
        ax.grid(which='major', axis='x', alpha=0.75)
        ax.grid(which='minor', axis='x', alpha=0.2)
        ax.set_xlim(np.min(wav), np.max(wav))
        return ax.get_figure(), ax


import unittest
import numpy as np

class MyTestCase(unittest.TestCase):
    def test_generate(self):
        from hylite.reference import genImage, randomSpectra
        im,A = genImage()
        self.assertTrue(im.ydim() == 512) # check shape
        self.assertTrue(np.isfinite(im.data).all() ) # check all finite
        self.assertFalse((im.data == 0).all()) # check all non-zero
if __name__ == '__main__':
    unittest.main()



"""
Fit and visualise individual hyperspectral features.
"""

import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import least_squares
from scipy.signal import argrelmin
from tqdm import tqdm

from gfit import gfit, initialise, evaluate

class HyFeature(object):
    """
    Utility class for representing and fitting individual or multiple absorption features.
    """

    def __init__(self, name, pos, width, depth=1, data=None, color='g'):
        """
        Args:
            name (str) a name for this feature.
            pos (float): the position of this feature (in nm).
            width (float): the width of this feature (in nm).
            data (ndarray): a real spectra associated with this feature (e.g. for feature fitting or from reference libraries).
                  Should be a numpy array such that data[0,:] gives wavelength and data[1,:] gives reflectance.
        """

        self.name = name
        self.pos = pos
        self.width = width
        self.depth = depth
        self.color = color
        self.data = data
        self.mae = -1
        self.strength = -1
        self.components = None
        self.endmembers = None

    def get_start(self):
        """
        Get start of feature.

        Returns:
            the feature position - 0.5 * feature width.
        """

        return self.pos - self.width * 0.5

    def get_end(self):
        """
        Get approximate end of feature

        Returns:
        returns feature position - 0.5 * feature width.
        """

        return self.pos + self.width * 0.5


    ######################
    ## Feature models
    ######################
    @classmethod
    def gaussian(cls, _x, pos, width, depth):
        """
        Static function for evaluating a gaussian feature model

        Args:
            x (ndarray): wavelengths (nanometres) to evaluate the feature over
            pos (float): position for the gaussian function (nanometres)
            width (float): width for the gaussian function.
            depth (float): depth for the gaussian function (max to min)
            offset (float): the vertical offset of the functions. Default is 1.0.
        """
        return 1 - depth * np.exp( -(_x - pos)**2 / width )

    @classmethod
    def multi_gauss(cls, x, pos, width, depth, asym=None):
        """
        Static function for evaluating a multi-gaussian feature model

        Args:
            x (ndarray): wavelengths (nanometres) to evaluate the feature over
            pos (list): a list of positions for each individual gaussian function (nanometres)
            width (list): a list of widths for each individual gaussian function.
            depth (list): a list of depths for each individual gaussian function (max to min)
            asym (list): a list of feature asymmetries. The right-hand width will be calculated as:
                         w2 = asym * width. Default is 1.0.
        """
        if asym is None:
            asym = np.ones( len(width) )
        M = np.hstack( [depth[i], pos[i], width[i], width[i]*asym[i]] for i in range(len(depth)) )
        y = evaluate( x, M, sym=False )
        return 1 - y

    # noinspection PyDefaultArgument
    def quick_plot(self, method='gauss', ax=None, label='top', lab_kwds={}, **kwds):
        """
        Quickly plot this feature.

        Args:
            method (str): the method used to represent this feature. Options are:

                        - 'gauss' = represent using a gaussian function
                        - 'range' = draw vertical lines at pos - width / 2 and pos + width / 2.
                        - 'fill' = fill a rectangle in the region dominated by the feature with 'color' specifed in kwds.
                        - 'line' = plot a (vertical) line at the position of this feature.
                        - 'all' = plot with all of the above methods.

            ax: an axis to add the plot to. If None (default) a new axis is created.
            label (float): Label this feature (using it's name?). Options are None (no label), 'top', 'middle' or 'lower'. Or,
                   if an integer is passed, odd integers will be plotted as 'top' and even integers as 'lower'.
            lab_kwds (dict): Dictionary of keywords to pass to plt.text( ... ) for controlling labels.
            **kwds: Keywords are passed to ax.axvline(...) if method=='range' or ax.plot(...) otherwise.

        Returns:
            Tuple containing

            - fig: the figure that was plotted to.
            - ax: the axis that was plotted to.
        """

        if ax is None:
            fig, ax = plt.subplots()

        # plot reference spectra and get _x for plotting
        if self.data is not None:
            _x = self.data[0, : ]
            ax.plot(_x, self.data[1, :], color='k', **kwds)
        else:
            _x = np.linspace(self.pos - self.width, self.pos + self.width)

        # set color
        if 'c' in kwds:
            kwds['color'] = kwds['c']
            del kwds['c']
        kwds['color'] = kwds.get('color', self.color)

        # get _x for plotting
        if 'range' in method.lower() or 'all' in method.lower():
            ax.axvline(self.pos - self.width / 2, **kwds)
            ax.axvline(self.pos + self.width / 2, **kwds)
        if 'line' in method.lower() or 'all' in method.lower():
            ax.axvline(self.pos, color='k', alpha=0.4)
        if 'gauss' in method.lower() or 'all' in method.lower():
            if self.components is None: # plot single feature
                _y = HyFeature.gaussian(_x, self.pos, self.width, self.depth)
            else:
                _y = HyFeature.multi_gauss(_x, [c.pos for c in self.components],
                                               [c.width for c in self.components],
                                               [c.depth for c in self.components] )
            ax.plot(_x, _y, **kwds)
        if 'fill' in method.lower() or 'all' in method.lower():
            kwds['alpha'] = kwds.get('alpha', 0.25)
            ax.axvspan(self.pos - self.width / 2, self.pos + self.width / 2, **kwds)

        # label
        if not label is None:

            # calculate label position
            rnge = ax.get_ylim()[1] - ax.get_ylim()[0]
            if isinstance(label, int):
                if label % 2 == 0:
                    label = 'top'  # even
                else:
                    label = 'low'  # odd
            if 'top' in label.lower():
                _y = ax.get_ylim()[1] - 0.05 * rnge
                va = lab_kwds.get('va', 'top')
            elif 'mid' in label.lower():
                _y = ax.get_ylim()[0] + 0.5 * rnge
                va = lab_kwds.get('va', 'center')
            elif 'low' in label.lower():
                _y = ax.get_ylim()[0] + 0.05 * rnge
                va = lab_kwds.get('va', 'bottom')
            else:
                assert False, "Error - invalid label position '%s'" % label.lower()

            # plot label
            lab_kwds['rotation'] = lab_kwds.get('rotation', 90)
            lab_kwds['alpha'] = lab_kwds.get('alpha', 0.5)
            ha = lab_kwds.get('ha', 'center')
            if 'ha' in lab_kwds: del lab_kwds['ha']
            if 'va' in lab_kwds: del lab_kwds['va']
            lab_kwds['bbox'] = lab_kwds.get('bbox', dict(boxstyle="round",
                                                         ec=(0.2, 0.2, 0.2),
                                                         fc=(1., 1., 1.),
                                                         ))
            ax.text(self.pos, _y, self.name, va=va, ha=ha, **lab_kwds)

        return ax.get_figure(), ax

class MultiFeature(HyFeature):
    """
    A spectral feature with variable position due to a solid solution between known end-members.
    """

    def __init__(self, name, endmembers):
        """
        Args:
            endmembers (list): a list of HyFeature objects representing each end-member.
        """

        # init this feature so that it ~ covers all of its 'sub-features'
        minw = min([e.pos - e.width / 2 for e in endmembers])
        maxw = max([e.pos + e.width / 2 for e in endmembers])
        depth = np.mean([e.depth for e in endmembers])
        super().__init__(name, pos=(minw + maxw) / 2, width=maxw - minw, depth=depth, color=endmembers[0].color)

        # store endmemebers
        self.endmembers = endmembers

    def count(self):
        return len(self.endmembers)

    def quick_plot(self, method='fill+line', ax=None, suplabel=None, sublabel=('alternate', {}), **kwds):
        """
         Quickly plot this feature.

         Args:
            method (str): the method used to represent this feature. Default is 'fill+line'. Options are:

                         - 'gauss' = represent using a gaussian function at each endmember.
                         - 'range' = draw vertical lines at pos - width / 2 and pos + width / 2.
                         - 'fill' = fill a rectangle in the region dominated by the feature with 'color' specifed in kwds.
                         - 'line' = plot a (vertical) line at the position of each feature.
                         - 'all' = plot with all of the above methods.

            ax: an axis to add the plot to. If None (default) a new axis is created.
            suplabel (str): Label positions for this feature. Default is None (no labels). Options are 'top', 'middle' or 'lower'.
            sublabel (str): Label positions for endmembers. Options are None (no labels), 'top', 'middle', 'lower' or 'alternate'. Or, if an integer
                    is passed then it will be used to initialise an alternating pattern (even = top, odd = lower).
            lab_kwds (dict): Dictionary of keywords to pass to plt.text( ... ) for controlling labels.
            **kwds: Keywords are passed to ax.axvline(...) if method=='range' or ax.plot(...) otherwise.

         Returns:
            Tuple containing

            - fig: the figure that was plotted to
            - ax: the axis that was plotted to
         """

        if ax is None:
            fig, ax = plt.subplots()

        # plot
        if 'range' in method.lower() or 'all' in method.lower():
            super().quick_plot(method='range', ax=ax, label=None, **kwds)
        if 'line' in method.lower() or 'all' in method.lower():
            for e in self.endmembers:  # plot line for each end-member
                e.quick_plot(method='line', ax=ax, label=None, **kwds)
        if 'gauss' in method.lower() or 'all' in method.lower():
            for e in self.endmembers:  # plot gaussian for each end-member
                e.quick_plot(method='gauss', ax=ax, label=None, **kwds)
                if isinstance(sublabel, int): sublabel += 1
        if 'fill' in method.lower() or 'all' in method.lower():
            super().quick_plot(method='fill', ax=ax, label=None, **kwds)

        # and do labels
        if not suplabel is None:
            if not isinstance(suplabel, tuple): suplabel = (suplabel, {})
            super().quick_plot(method='label', ax=ax, label=suplabel[0], lab_kwds=suplabel[1])
        if not sublabel is None:
            if not isinstance(sublabel, tuple): sublabel = (sublabel, {})
            if isinstance(sublabel[0], str) and 'alt' in sublabel[0].lower():
                sublabel = (1, sublabel[1])  # alternate labelling
            for e in self.endmembers:
                e.quick_plot(method='label', ax=ax, label=sublabel[0], lab_kwds=sublabel[1])
                sublabel = (sublabel[0] + 1, sublabel[1])
        return ax.get_figure(), ax

class MixedFeature(HyFeature):
    """
    A spectral feature resulting from a mixture of known sub-features.
    """

    def __init__(self, name, components, **kwds):
        """
        Args:
            components: a list of HyFeature objects representing each end-member.
            **kwds: keywords are passed to HyFeature.init()
        """

        # init this feature so that it ~ covers all of its 'sub-features'
        minw = min([e.pos - e.width / 2 for e in components])
        maxw = max([e.pos + e.width / 2 for e in components])
        depth = np.mean([e.depth for e in components])

        if not 'color' in kwds:
            kwds['color'] = components[0].color
        super().__init__(name, pos=(minw + maxw) / 2, width=maxw - minw, depth=depth, **kwds)

        # store components
        self.components = components

    def count(self):
        return len(self.components)


from abc import ABCMeta, abstractmethod
import os
import numpy as np
import hylite.io as io
from hylite.project import Camera

class Sensor(object):
    """
    Base class for hyperspectral camera specific implementations to inherit. This defines the basic functions that
    every sensor class is expected to implement.
    """

    __metaclass__ = ABCMeta

    #set all default references to none
    dark = None # dark reference image
    white = None # white reference image
    white_spectra = None # reflectance of white (or grey) target

    @abstractmethod
    def name(cls):
        """
        Returns this sensors name
        """

        pass #implement this in inherited classes

    @classmethod
    def focal_length(cls):
        """
        Return the sensor focal length (in pixels).
        """
        return cls.ypixels() / (2 * np.tan(np.deg2rad(cls.fov() / 2)))

    @abstractmethod
    def fov(cls):
        """
        Return the (vertical) sensor field of view .
        """
        pass # implement this in inherited classes

    @abstractmethod
    def ypixels(cls):
        """
        Return the number of pixels in the y-dimension.
        """
        pass # implement this in inherited classes

    @abstractmethod
    def xpixels(cls):
        """
        Return the number of pixels in the x-dimension (==1 for line scanners).
        """
        pass # implement this in inherited classes

    @abstractmethod
    def pitch(cls):
        """
        Return the pitch of the each pixel in the x-dimension (though most pixels are square). Note that for
        line scanners this will be an angular pitch.
        """
        pass # implement this in inherited classes

    @classmethod
    def correct_image(cls, image, verbose=False, **kwds):
        """
        Apply sensor corrections to an image.

        Args:
            image (hylite.HyImage): a hyImage instance of an image captured using this sensor.
            verbose (bool): true if updates/progress should be printed to the console. Default is False.
            **kwds: keywords as defined by inherited classes (sensor specific)
        """

        pass  # implement this in inherited classes

    @abstractmethod
    def correct_folder(cls, path ):
        """
        Many sensors use simple/common data structures to store data/headers/dark reference etc. Hence it is often easiest
        to pass an output folder to the sensor for correction.

        Args:
            path (str): a path to the folder containing the sensor specific data.
            **kwds: keywords as defined by inherited classes (sensor specific)

        Returns:
            A hyImage to which all sensor-specific corrections have been applied. Note that this will generally not include
            topographic or atmospheric corrections.

        """

        pass  # implement this in inherited classes

    @classmethod
    def set_dark_ref(cls, image):
        """
        Sets the dark reference to be used for sensor corrections.
        """

        if isinstance(image, str):
            assert os.path.exists(image), "Error: %s is not a valid file path or hyperspectral image."
            image = io.load(image)
        assert isinstance(image, io.HyImage) or image is None, "Error: dark reference must be an image or None."
        Sensor.dark = image

    @classmethod
    def set_white_ref_spectra(cls, spectra):
        """
        Set the known reflectance spectra of the white reference.

        Args:
            spectra (hylite.reference.spectra.Target): The known reflectance spectra of the white reference.
                     If None, then the target is assumed to be pure white (reflectance = 1).
        """
        Sensor.white_spectra = spectra # store reference spectra

    @classmethod
    def set_white_ref(cls, image):
        """
        Sets the white reference to be used for sensor corrections.

        Args:
            image (hylite.HyImage): the white reference image.
        """

        if isinstance(image, str):
            assert os.path.exists(image), "Error: %s is not a valid file path or hyperspectral image."
            image = io.load(image)

        assert isinstance(image, io.HyImage), "Error: white reference must be an image."
        Sensor.white = image # store white reference

    @classmethod
    def get_camera(cls):
        """
        Return a hylite.project.Camera object with fov, pitch and projection corresponding to this sensor.
        """
        if cls.xpixels() == 1: # is this a line-scanner?
            return Camera(np.zeros(3), np.zeros(3), 'pano', cls.fov(), (cls.xpixels(), cls.ypixels()), cls.pitch())
        else:
            return Camera(np.zeros(3), np.zeros(3), 'pano', cls.fov(), (cls.xpixels(), cls.ypixels()) )



import unittest
import hylite
from hylite.project import Camera
from hylite.reference.spectra import R90
from hylite.correct.panel import Panel

import numpy as np
from tests import genHeader, genCloud, genImage

class TestHyData(unittest.TestCase):
    def test_header(self):

        #load header from file
        header = genHeader()

        # check basics
        self.assertEqual(header.has_band_names(), False)
        self.assertEqual(header.has_wavelengths(), True)
        self.assertEqual(header.has_fwhm(), False)
        self.assertEqual(header.band_count(), 450)
        self.assertEqual(len(header.get_wavelengths()), 450)

        # check copy and set functions
        header2 = header.copy()
        header2.set_wavelengths( np.zeros_like(header.get_wavelengths()))
        header2.set_band_names(["Band %d" for i in range(header.band_count())])
        self.assertEqual( (header.get_wavelengths() == header2.get_wavelengths()[0]).any(), False )

        # check drop bands
        header3 = header.copy()
        header3.set_band_names(["Band %d" for i in range(header.band_count())])
        mask = np.full( header2.band_count(), True )
        mask[0:4] = False
        header3.drop_bands(mask)
        self.assertEqual(header3.band_count(), 4)
        self.assertEqual(len(header3.get_wavelengths()), 4)
        self.assertEqual(len(header3.get_band_names()), 4)

        # check set Camera
        # define camera properties and initial location/orientation estimate
        cam = Camera(pos=np.asarray([665875.0, 4162695, 272]),  # np.array([666290.454, 4162697.93, 268.521235])
                     ori=np.array([43, 80, 130]),  # np.array([50.0,-83.0,-137.0])
                     proj='pano', fov=32.3, step=0.084,
                     dims=(1464, 401))
        header.set_camera(cam)
        cam2 = header.get_camera()

        self.assertEqual((cam2.pos == cam.pos).all(), True)
        self.assertEqual((cam2.ori == cam.ori).all(), True)
        self.assertEqual(cam2.proj, cam.proj)
        self.assertEqual(cam2.dims, cam.dims)
        self.assertEqual(cam2.fov, cam.fov)
        self.assertEqual(cam2.step, cam.step)

        # check set panel
        panel = Panel( R90, np.zeros( header.band_count() ), wavelengths=header.get_wavelengths() )
        header.add_panel(panel)
        self.assertEqual( len(header.get_panel_names()), 1)
        panel2 = header.get_panel('R90')
        self.assertEqual( np.sum( panel2.get_mean_radiance() ), 0 )
        self.assertEqual(panel2.get_mean_radiance().shape[0], header.band_count())
        self.assertEqual( panel2.material.get_name().lower(), R90.get_name().lower())

    def test_data(self):
        # check functions for images and cloud data
        lines = [401, 1]
        samples = [1464, 1000]
        for i,data in enumerate( [genImage(dimx = 1464, dimy=401, nbands=10), genCloud(npoints = 1000, nbands=10)] ):
            # check basics
            self.assertEqual(data.has_wavelengths(), True)
            self.assertEqual(data.has_band_names(), True)
            self.assertEqual(data.has_fwhm(), True)
            self.assertEqual(data.band_count(), 10)
            self.assertEqual(data.samples(), samples[i])
            self.assertEqual(data.lines(), lines[i])
            self.assertEqual(data.is_int(), False)
            self.assertEqual(data.is_float(), True)

            # check band names
            data.set_band_names([a for a in 'abcdefghijklmnop'[:10]])
            self.assertEqual( data.get_band_index('e'), 4 )

            # check export (which also checks copy etc.)
            data2 = data.export_bands( (0,5) )
            self.assertEqual(len(data2.get_wavelengths()), 6)
            self.assertEqual(len(data2.get_fwhm()), 6)
            self.assertEqual(data2.data.shape[-1], 6 )

            # nans
            data2.mask_bands(3,-1) # mask bands from 3rd to last
            self.assertEqual(data2.data.shape[-1], 6) # bands should still exist
            self.assertEqual( np.isfinite(data2.data[...,3:]).any(), False ) # all of last bands should be nan
            data2.delete_nan_bands()
            self.assertEqual(data2.data.shape[-1], 3)  # bands should have been deleted
            self.assertEqual(len(data2.get_wavelengths()), 3) # as should associated header data
            self.assertEqual(len(data2.get_fwhm()), 3) # as should associated header data

            # test set as nan
            self.assertEqual(np.isfinite(data2.data[..., 2]).all(), True)
            data2.data[..., :] = 0
            data2.set_as_nan(0)
            self.assertEqual( np.isfinite( data2.data[...,2] ).all(), False )

            # get band
            self.assertEqual( np.isfinite( data2.get_band(2)).any(), False)
            self.assertEqual( data2.get_band_grey(0).dtype, np.uint8 )
            self.assertEqual( data2.get_band_index(500.0), 0 )

            # check compression
            tv = data.data.ravel()[0]
            data.compress()
            self.assertEqual(data.data.dtype, np.uint16)
            data.decompress()
            self.assertEqual(data.data.dtype, np.float32)
            self.assertAlmostEqual(data.data.ravel()[0], tv, 3)

            # check quantize
            for m in ['kmeans']: # , 'minibatch', 'birch']:
                index,lib = data.getQuantized(n=255, cmeth=m, vthresh=10, subsample=50, mask=None )
                self.assertEquals( np.max(index.data), 255 )
                self.assertEquals( lib.data.shape[0], 256 )

            # check reconstruction from quanta
            rc = hylite.HyData.fromQuanta( index, lib )
            self.assertTrue( (np.array(rc.data.shape) == np.array(data.data.shape)).all() )
            self.assertListEqual( list(rc.get_wavelengths()), list(data.get_wavelengths()) )
            self.assertListEqual( list(rc.get_band_names()), list(data.get_band_names()) )

            # check smoothing works with nan bands
            data.mask_bands(1, 3)
            data.mask_bands(8, -1)
            data.smooth_median(window=3)
            data.smooth_savgol(window=3, chunk=True)

            # percent clip
            data.percent_clip(5,95,per_band=False,clip=True)
            data.percent_clip(5, 95, per_band=True, clip=True)

            # expression evaluation
            out = data.eval('b2:b5 + b8')
            w0 = data.get_wavelengths()[1]
            w1 = data.get_wavelengths()[3]
            self.assertEqual(out.data.shape[-1], 1 ) # should have 1 band
            out = data.eval('b2:b5 + b8 | %d/(%d+$1)**$2'%(w0, w1))
            self.assertEqual(out.data.shape[-1], 2 ) # should have 2 bands

            # normalise
            data.normalise()

            # resampling
            sub = data.resample( data.get_wavelengths()[2::4], agg = True, thresh=30. )
            self.assertEqual( sub.band_count(), int( data.band_count() / 4 ) )
            sub2 = data.resample(data.get_wavelengths()[2::5], agg=False, thresh=30.)
            self.assertEqual(sub2.band_count(), int(data.band_count() / 5))
            rg = [ (data.get_wavelengths()[i], data.get_wavelengths()[i+2]) for i in range(0,data.band_count()-2, 2) ]
            sub3 = data.resample(rg, agg=True, thresh=0.1)
            self.assertEqual(sub3.band_count(), len(rg))
            sub4 = data.resample( np.linspace(0.,1000.), partial=True)

            # test fill holes
            data.data = np.nan_to_num(data.data, posinf=0, neginf=0) # remove any stray nans
            self.assertTrue(np.isfinite(data.data).all())
            data.data[...,5] = np.nan # add some nans
            data.fill_gaps() # remove them again!
            self.assertTrue(np.isfinite(data.data).all() ) # check they were removed

            # test reshaping to feature vectors
            data.data[..., 5] = np.nan  # add some nans
            self.assertSequenceEqual( data.X().shape, np.reshape(data.data, (-1, data.band_count())).shape ) # check shape is unchanged
            self.assertNotEquals(data.X(True).shape[0], np.reshape(data.data, (-1, data.band_count())).shape[0] ) # check nans are removed
            self.assertTrue( np.isfinite( data.X(True).all()) )

            data.set_raveled( np.zeros_like( data.X(True)), onlyFinite=True ) # set with nan-mask
            self.assertFalse( np.isfinite(data.data ).all() ) # check nans persist

            data.set_raveled(np.zeros_like(data.X()), onlyFinite=False)  # set without nan-mask
            self.assertTrue(np.isfinite(data.data).all())  # check nans are gone

            # N.B. data.data is now all zero!

if __name__ == '__main__':
    unittest.main()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="hylite",
    version="1.30",
    author="Helmholtz Institute Freiberg",
    author_email="s.thiele@hzdr.de",
    description="Open-source toolbox for hyperspectral geology.",
    long_description="A python package for loading, correcting, projecting and analysing hyperspectral datasets, with particular emphasis on geological applications.",
    long_description_content_type="text/markdown",
    url="https://github.com/samthiele/hylite",
    #package_dir={'': 'hylite'},
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering"
    ],
    keywords='hyperspectral data analysis hypercloud geology mineral mapping',
    python_requires='>=3.6',
    install_requires=["scipy>=1.4", "matplotlib>=3", "numpy",
                      "imageio","opencv-python>=4.5", "opencv-contrib-python>=4.5",
                      "scikit-image", "tqdm", "roipoly", "spectral", "utm","pytz",
                      "laspy","plyfile","astral","piexif","gfit","numba","natsort"],

    # List additional URLs that are relevant to your project as a dict.
    #
    # This field corresponds to the "Project-URL" metadata fields:
    # https://packaging.python.org/specifications/core-metadata/#project-url-multiple-use
    #
    # Examples listed include a pattern for specifying where the package tracks
    # issues, where the source is hosted, where to say thanks to the package
    # maintainers, and where to support the project financially. The key is
    # what's used to render the link text on PyPI.
    project_urls={  # Optional
        'Source': 'https://github.com/samthiele/hylite',
        'Documentation' : 'https://hifexplo.github.io/hylite/hylite.html',
    },

    package_data={"": ["*.txt","*.hdr","*.cal","*.dat"] }
)



import unittest
from hylite import io
from pathlib import Path
import numpy as np
import os

class MyTestCase(unittest.TestCase):
    def test_sam(self):
        from hylite.analyse.sam import SAM, spectral_angles

        image = io.load(os.path.join(os.path.join(str(Path(__file__).parent.parent), "test_data"),"image.hdr"))
        cloud = io.load(os.path.join(os.path.join(str(Path(__file__).parent.parent), "test_data"),"image.hdr"))

        for data in [image, cloud]:
            em1 = data.X(onlyFinite=True)[0]
            em2 = data.X(onlyFinite=True)[-1]

            # check spectral angles function runs
            ang = spectral_angles( [em1,em2], data.X() )
            self.assertTrue(np.isfinite(ang).any())

            # run SAM
            em3 = data.X(onlyFinite=True)[-10]
            sam = SAM(data, [[em3], [em1,em2] ])
            self.assertTrue(np.isfinite(sam.data).any())
            self.assertEquals( int( sam.X( onlyFinite=True )[0,0]), 1 )
            self.assertEquals(int(sam.X(onlyFinite=True)[-1, 0]), 1 )
            self.assertEquals(int(sam.X(onlyFinite=True)[-10, 0]), 0 )



"""
Functions for modelling the amount of incident light that is reflected towards the camera, including Lambertian
and Oren-Nayar BRDFs.
"""

import numpy as np
import hylite

def estimate_incidence(normals, sunvec):
    """
    Utility function to estimate the cosine of incidence angles based on normals and calculated sun position.

    Args:
        normals: either: (1) a HyImage with band 0 = nx, band 1 = ny and band 2 = nz, (2) HyCloud instance containing
                 normals, or (3) mx3 numpy array of normal vectors.
        sunvec: a numpy array containing the sun illumination vector (as calculated by estimate_sun_vec(...)).

    Returns:
        list of incidence angles matching the shape of input data (but with a single band only).
    """

    # extract normal vectors
    if isinstance(normals, hylite.HyCloud):
        N = normals.normals[:, :3]
        outshape = normals.point_count()
    elif isinstance(normals, hylite.HyImage):
        N = normals.get_raveled()[:, :3]
        outshape = (normals.xdim(), normals.ydim())
    else:
        N = normals.reshape((-1, 3))
        outshape = normals.shape[:-1]

    # normalize normals (just to be safe)
    N = N / np.linalg.norm(N, axis=1)[:, None]

    # calculate cosine of angles used in correction
    cosInc = np.dot(-N, sunvec)  # cos incidence angle

    # return in same shape as original data
    return cosInc.reshape(outshape)

def calcOrenNayar(normals, view, source, roughness=0.3):
    """
    Calculate OrenNayar reflectance across a scene.

    Args:
        normals: a (...,3) array containing normal vectors for each point or pixel.
        view: a (...,3) array containing viewing directions for each point or pixel.
        source: a (3,) array containing the illumination direction.
        roughness: a float describing the OrenNayar roughness parameter. Default is 0.3.

    Returns:
        a (...,1) numpy array containing the Oren-Nayar reflectance factors.
    """

    # check array shapes are compatible
    assert normals.shape == view.shape, "Error - normals array and view array must have the same shape."
    assert normals.shape[-1] == 3, "Error - the last axis of the normal and view arrays must have 3 values (x,y,z)."
    assert source.shape[0] == 3 and len(source.shape) == 1, "Error - source must be a 3-D vector."

    outshape = normals.shape[:-1] + (1,)  # output shape
    N = normals.reshape((-1, 3))  # flatten
    V = view.reshape((-1, 3))
    r2 = roughness ** 2
    I = source  # shorthand for later

    # calculate roughness terms
    A = 1.0 - (0.5 * r2) / (r2 + 0.33)
    B = (0.45 * r2) / (r2 + 0.09)

    LdotN = estimate_incidence(N, I)  # I . N [ = cos (incidence angle) ]
    VdotN = np.sum(V * N, axis=-1)  # V . N [ = cos( viewing angle ) ]

    # remove backfaces
    irradiance = LdotN.copy()
    irradiance[irradiance < 0] = 0

    # convert cosines to radians
    angleViewNormal = np.arccos(VdotN);
    angleLightNormal = np.arccos(LdotN);

    a = (V - N * VdotN[:, None]) / np.linalg.norm((V - N * VdotN[:, None]), axis=-1)[:, None]
    b = (I - N * LdotN[:, None]) / np.linalg.norm((I - N * LdotN[:, None]), axis=-1)[:, None]

    angleDiff = np.sum(a * b, axis=-1)
    angleDiff[angleDiff < 0] = 0

    alpha = np.nanmax(np.dstack([angleViewNormal, angleLightNormal]), axis=-1)[0, :]
    beta = np.nanmin(np.dstack([angleViewNormal, angleLightNormal]), axis=-1)[0, :]

    # return
    return (irradiance * (A + B * angleDiff * np.sin(alpha) * np.tan(beta))).reshape(outshape)

def calcLambert(normals, source):
    """
    Calculate lambertian reflectance across a scene.

    Args:
        normals: a (...,3) array containing normal vectors for each point or pixel.
        source: a (3,) array containing the illumination direction.

    Returns:
        a (...,1) numpy array containing the Lambertian reflectance factors.
    """

    # check array shapes are compatible
    assert normals.shape[-1] == 3, "Error - the last axis of the normal and view arrays must have 3 values (x,y,z)."
    assert source.shape[0] == 3 and len(source.shape) == 1, "Error - source must be a 3-D vector."

    outshape = normals.shape[:-1] + (1,)  # output shape

    # I . N [ = cos (incidence angle) ]
    return np.clip(estimate_incidence(normals.reshape((-1, 3)), source), 0, 1).reshape(outshape)




"""
Use spectra derivatives to extract and analyse turning points (maxima and minima).
"""

import numpy as np
from tqdm import tqdm
from scipy import signal


import warnings
warnings.filterwarnings("ignore", category=np.VisibleDeprecationWarning)

def build_kernel(sigma, res):
    """
    Construct a kernel (normal distribution with specified standard deviation)
    """
    kx = np.arange(-5 * sigma, 5 * sigma, step=res, dtype=np.float32)
    ky = ((1 / (sigma * np.sqrt(2 * np.pi))) * np.exp(-0.5 * ((kx / sigma) ** 2))).astype(np.float32)
    ky /= np.trapz(ky, kx)  # increase values slightly to account for numerical integration errors
    return ky


def TPT(data, sigma=10., window=21, n=2, thresh=0, domain=None, weighted=True, maxima=True, minima=True, vb=True):
    """
    Extract the turning points from each spectra in dataset, and build an associated kernel density estimates that
    can be used to calculate the probability of observing given features.

    Args:
        data: a HyData instance containing the spectra to filter.
        sigma: Standard deviation of the KDE kernel. Either a number, or a function such that sigma(w) returns the
               standard deviation for a specific wavelength (e.g. if a larger kernel should be used for
               LWIR than VNIR data). Default is 10 nm.
        window: the size of the window (in indices) used during savgol smoothing and derivative calculation. Default is 21.
        n: the order of the polynomial to use for savgol smoothing. Larger number preserve smaller features but
           are slower to compute (and more prone to noise). Default is 2.
        thresh: a depth/prominence threshold to ignore small maxima or minima. Default is 0.
        domain: None (default) if the KDE should be evaluated over the same wavelengths as the input data. Alternatively, a
                tuple can be passed containing (min_wav, max_wav, [resolution (optional)]).
        weighted: True if the KDE should be weighted based on feature depth (i.e. more prominent features get extra
                  weight in the KDE as they are ~ 'more likely to be observed'). Default is True.
        maxima: True the filter should include maxima. Default is True.
        minima: True if the filter should include minima. Default is True.
        vb: True progress should be printed to the console. Default is True.

    Returns:
        A tuple containing:

         - tpt = a HyData instance containg kernel density estimates of maxima and minima in the each pixel/point. Maxima
                 are given positive weight while minima are given negative weight.
         - Tpos = an array containing a list of turning point positions (wavelength) for each point/pixel.
         - Tdepth = an array containing a list of turning point depths for each point/pixel. Maxima are positive, minima negative.
    """

    # init wavelength domain
    if domain is not None:
        assert isinstance(domain, tuple) and (len(domain) == 3 or len(
            domain) == 2), "Domain must be a tuple of length two or three containing (start_wavelength, end_wavelength, resolution )."
        if (len(domain) == 3):
            w = np.arange(domain[0], domain[1], step=domain[2])
            res = domain[2]
        data = data.export_bands((domain[0], domain[1]))  # subset data
        if (len(domain) == 2):
            w = data.get_wavelengths()
            res = w[1] - w[0]
    else:
        w = data.get_wavelengths()
        res = w[1] - w[0]

    # get flattened reflectance array
    R = data.X()
    assert R.shape[-1] > window, "Error - window > band count. Please reduce window size (or increase band count)."

    # calculate derivatives
    mask = np.isfinite(R).all(axis=-1)
    dy = np.full( R.data.shape, np.nan )
    dy[mask,:] = signal.savgol_filter(R[mask, :], deriv=1, window_length=window, polyorder=n, axis=-1 )

    # init output array
    out = np.zeros((dy.shape[0], w.shape[0]))
    Tpos = [[] for i in range(dy.shape[0])]
    Tdepth = [[] for i in range(dy.shape[0])]
    K = {}
    loop = range(dy.shape[0])
    if vb:
        loop = tqdm(loop, leave=False, desc="Finding turning points")
    for p in loop:
        # find turning points
        minx = np.argwhere((dy[p, 1:] > 0) & (dy[p, :-1] < 0))[:, 0]
        maxx = np.argwhere((dy[p, 1:] < 0) & (dy[p, :-1] > 0))[:, 0]

        # add minima to output
        if minima:
            for i in minx:
                # match feature with adjacent maxima
                left = 0
                right = dy.shape[-1] - 1
                if (maxx < i).any():
                    left = maxx[maxx < i][np.argmin(np.abs(i - maxx[maxx < i]))]
                if (maxx > i).any():
                    right = maxx[maxx > i][np.argmin(np.abs(maxx[maxx > i] - i))]

                # calculate depth and associated weight
                d = max(R[p, left], R[p, right]) - R[p, i]  # compute simple maxima - minima depth
                weight = 1.0
                if weighted:
                    weight = d / max(R[p, left],
                                     R[p, right])  # weight is the fractional decrease associated with this minima

                # calculate index in output array
                if domain is not None:
                    idx = int((data.get_wavelengths()[i] - w[0]) / res)  # compute idx in ouput array
                else:
                    idx = i  # no difference

                # calculate/get kernel
                if callable(sigma):
                    s = sigma(w[idx])  # get sigma for this wavelength
                else:
                    s = sigma  # use fixed sigma
                if s not in K:  # build kernel if not already done
                    K[s] = build_kernel(s, res)

                    # add to output
                if d > thresh:
                    # store position
                    Tpos[p].append(w[idx])
                    Tdepth[p].append(-d)

                    # apply kernel
                    k0 = int(idx - (K[s].shape[0] / 2))  # index of start of kernel (pos - 5 sigma)
                    k1 = k0 + K[s].shape[0]  # index of end of kernel (pos + 5 sigma)
                    if k1 < out.shape[
                        -1] and k0 > 0:  # ignore features close to edges (these are probably dodgy anyway)
                        out[p, k0:k1] -= weight * K[s]  # add kernel to likelihood function

        # add maxima to output
        if maxima:
            for i in maxx:
                # match feature with adjacent maxima
                left = 0
                right = dy.shape[-1] - 1
                if (minx < i).any():
                    left = minx[minx < i][np.argmin(np.abs(i - minx[minx < i]))]
                if (minx > i).any():
                    right = minx[minx > i][np.argmin(np.abs(minx[minx > i] - i))]

                # calculate height
                d = R[p, i] - min(R[p, left], R[p, right])  # compute simple max - min height
                weight = 1.0
                if weighted:
                    weight = d / min(R[p, left],
                                     R[p, right])  # weight is the fractional increaes associated with this maxima

                # calculate index in output array
                if domain is not None:
                    idx = int((data.get_wavelengths()[i] - w[0]) / res)  # compute idx in ouput array
                else:
                    idx = i  # no difference

                # calculate/get kernel
                if callable(sigma):
                    s = sigma(w[idx])  # get sigma for this wavelength
                else:
                    s = sigma  # use fixed sigma
                if s not in K:  # build kernel if not already done
                    K[s] = build_kernel(s, res)

                    # add kernel to output
                if d > thresh:
                    # store position
                    Tpos[p].append(w[idx])
                    Tdepth[p].append(d)

                    # calculate position (index)
                    k0 = int(idx - (K[s].shape[0] / 2))  # index of start of kernel (pos - 5 sigma)
                    k1 = k0 + K[s].shape[0]  # index of end of kernel (pos + 5 sigma)
                    if k1 < out.shape[
                        -1] and k0 > 0:  # ignore features close to edges (these are probably dodgy anyway)
                        out[p, k0:k1] += weight * K[s]  # add kernel to likelihood function

    # construct outputs
    O = data.copy(data=False)
    if data.is_image():  # reshape outputs
        Tpos = np.array(Tpos).reshape(data.data.shape[:-1])
        Tdepth = np.array(Tdepth).reshape(data.data.shape[:-1])
        O.data = out.reshape(data.data.shape[:-1] + (out.shape[-1],))
    else:
        O.data = out
    O.set_wavelengths(w)
    return O, Tpos, Tdepth


def TPT2MWL(pos, depth, wmin=0, wmax=-1, data=None, vb=True):
    """
    Convert the results from a turning point filter (TPT) into a minimum wavelength array. This will return
    the deepest feature within the specified range.

    Args:
        pos: an array containing per-spectra turning point positions, as returned by TPT(...).
        depth: an array containing per-spectra turning point depths, as returned by TPT(...).
        wmin: the start of the wavelength range of interest. Default is 0 (all wavelengths).
        wmax: the end of the wavelength range of interest. Default is -1 (all wavelengths).
        data: a HyData instance to use as a template for outputs. If not None (default) then this will return a HyData instance.
        vb: True if a progress bar should be created. Default is True.

    Returns:
        a numpy array or HyData instance (if data is not none), containing 4 bands (pos, width, depth, strength).
    """

    # reshape for loop
    shape = np.array(pos).shape
    pos = np.array(pos).ravel()
    depth = np.array(depth).ravel()
    out = np.full((len(pos), 4), np.nan, dtype=np.float32)

    # set wmax if need be
    if wmax == -1:
        wmax = np.inf

    # loop through spectra
    loop = range(len(pos))
    if vb:
        loop = tqdm(loop, leave=False, desc="Filtering turning points")
    for i in loop:
        if len(pos[i]) > 0:
            p = np.array(pos[i])
            d = np.array(depth[i])
            mask = (d < 0) & (p > wmin) & (p < wmax)
            if mask.any():
                idx = np.argmin(d[mask])
                out[i, 0] = p[mask][idx]
                out[i, 1] = 0  # how to estimate width?
                out[i, 2] = -d[mask][idx]
                out[i, 3] = out[i, 2]

    # return array for mwl map
    out = out.reshape(shape + (4,))
    if data is None:  # return a numpy array directly
        return out
    else:  # make a copy of data and return HyData instance
        data = data.copy(data=False)
        data.data = out
        return data


"""
Utility functions for sampling from hyperspectral datasets, including extracting averaged spectra (spectra libraries)
and resampling images to match band widths of other sensors (e.g. ASTER bands). CAUTION: this code is poorly tested.
"""

from hylite import HyData, HyLibrary
import numpy as np

def getClassLibrary( data, classification, skip0=True ):
    """
    Create a spectral library from a classified image by extracting all spectra associated with each class.

    Args:
        data: a HyImage or HyCloud instance containing spectral data.
        classification: a HyImage or HyCloud instance containing classifications per point/pixel.
        skip0: True if class 0 (typically background) should be ignored. Default is True.
    """

    assert isinstance(data, HyData), "Error - data must be a HyData instance not %s" % type(data)
    assert isinstance(classification, HyData), "Error - classification must be a HyData instance not %s" % type(classification)
    if not classification.is_classification():
        print( "Warning - classification is not a classification dataset (at least according to its header)?")

    spectra = data.get_raveled()
    cls = classification.get_raveled()[:,0]
    names = classification.header.get_class_names()
    if len(names) == 0:
        names = [str(i) for i in range(np.max(cls))]

    refl = []
    upper = []
    lower = []
    n = []
    mn = 0
    if skip0:
        mn = 1
    for i in range(mn,len(names)):
        msk = cls==i
        if msk.any():
            l, m, u = np.nanpercentile(spectra[msk,:], (25, 50, 75), axis=0)
            lower.append(l)
            refl.append(m)
            upper.append(u)
            n.append(names[i])

    return HyLibrary(n, np.vstack(refl),
                     lower=np.vstack(lower),
                     upper=np.vstack(upper),
                     wav=data.get_wavelengths())

class Resample( object ):
    """
    A wrapper class for spectral resampling of hyperspectral images.
    """

    def __init__( self, bands ):
        """
        Args:
            bands (list): a list of wavelength tuples specifying the minimum and maximum spectra of each band in this resampling scheme.
        """

        # check valid
        assert len(bands) > 0, "Error - no valid bands specified"
        for b in bands:
            assert len(b) == 2, "Error - bands must be specified as a tuple of length 2."
        self.bands = bands

    def get_band(self, data, n ):
        """
        Get the n'th band under this resampling scheme by averaging hyperspectral bands between the
        specified range.

        Args:
            data: the dataset to extract information from.
            n: the resampled band index to extract. NOTE THAT BAND INDICES START AT 1 FOR COMPATIBILITY WITH
               STANDARD SATELLITE NOTATION!
        """
        assert n >= 1 and (n-1) < len(self.bands), "Error - Band %d is not defined in this resampling scheme." % n
        idx0 = data.get_band_index( self.bands[n-1][0], thresh=np.inf )
        idx1 = data.get_band_index( self.bands[n-1][1], thresh=np.inf )

        if idx1 != idx0:
            return np.nanmean( data.data[...,idx0:idx1], axis=-1 )
        else:
            # is data within sampling range at all?
            minw,maxw = data.get_wavelengths()[[0,-1]]
            if (minw > self.bands[n-1][0]) and (maxw > self.bands[n-1][0]) \
                or (minw < self.bands[n-1][0]) and (maxw < self.bands[n-1][0]):
                return np.full( data.data.shape[:-1], np.nan ) # no data
            else:
                return data.data[..., idx0] # return this band

    def print_bands(self):
        """
        Quickly print the wavelengths associated with each band of this resampler.
        """

        for i,b in enumerate(self.bands):
            print("Band %d: %.1f - %.1f nm" % (i+1, b[0], b[1]))

    def apply(self, data):
        """
        Apply this resampling to a HyData instance and return a new instance with appropriately averaged bands.

        Args:
            data: the HyData instance to apply this sampling scheme to.
        Returns: A copy of the original HyData instance with the bands averaged as defined. Corresponding wavelengths will be set to the middle of each averaged region.
        """

        bands = [self.get_band(data, n + 1) for n in range(len(self.bands))]
        out = data.copy(data=False)
        if data.is_image():
            out.data = np.dstack(bands)
        else:
            out.data = np.hstack(bands)
        out.set_wavelengths([np.mean(self.bands[n]) for n in range(len(self.bands))])
        return out

# create instances for common satellites
# based on https://www.indexdatabase.de/db/bs.php

ASTER = Resample( [
    (520.0, 600.0),
    (630.0, 690.0),
    (760.0, 860.0),
    (1600.0, 1700.0),
    (2145.0, 2185.0),
    (2185.0, 2225.0),
    (2235.0, 2365.0),
    (2295.0, 2365.0),
    (2360.0, 2430.0),
    (8125.0, 8475.0),
    (8475.0, 8825.0),
    (8925.0, 9275.0),
    (10250.0, 10950.0),
    (10950.0, 11650.0) ] )
""" Static resampling class for sampling hyperspectral data onto ASTER bands"""

SENTINEL = Resample( [
    (433.0, 453.0),
    (458.0, 522.0),
    (543.0, 577.0),
    (650.0, 680.0),
    (698.0, 712.0),
    (733.0, 747.0),
    (773.0, 793.0),
    (785.0, 899.0),
    (855.0, 875.0),
    (935.0, 955.0),
    (1360.0, 1390.0),
    (1565.0, 1655.0),
    (2100.0, 2280.0) ] )
""" Static resampling class for sampling hyperspectral data onto Sentinel bands"""



import numpy as np
from pathlib import Path
import hylite.io as io
from hylite.sensors import Sensor
from scipy.ndimage import median_filter


class OWL(Sensor):
    """
    Implementation of sensor corrections for the OWL sensor.
    """

    @classmethod
    def name(cls):
        """
        Returns this sensors name
        """
        return "OWL"

    @classmethod
    def fov(cls):
        """
        Return the (vertical) sensor field of view .
        """
        return 32.3

    @classmethod
    def ypixels(cls):
        """
        Return the number of pixels in the y-dimension.
        """
        return 640  # n.b. sensor has 384 pixels but this is resized to 401 on lens correction

    @classmethod
    def xpixels(cls):
        """
        Return the number of pixels in the x-dimension (==1 for line scanners).
        """
        return 1  # the OWL is a line-scanner

    @classmethod
    def pitch(cls):
        """
        Return the pitch of the each pixel in the y-dimension (though most pixels are square).
        """
        return 0.05

    @classmethod
    def correct_image(cls, image, verbose=True, **kwds):

        """
        Apply sensor corrections to an image.

        Args:
            image (hylite.HyImage): a hyImage instance of an image captured using this sensor.
            verbose (bool): true if updates/progress should be printed to the console. Default is False.
            **kwds: Optional keywords include:

                 - rad = true if image should be converted to radiance by applying dark and white references. Default is True.
                 - bpr = replace bad pixels (only for raw data). Default is True.
                 - flip = true if image should be flipped (if camera mounted backwards in core
                          scanner). Default is False.
        """

        # get kwds
        rad = kwds.get("rad", True)
        bpr = kwds.get("bpr", True)

        lim = image.get_band_index(12368.)
        image.data = image.data[..., :lim]
        image.set_wavelengths(image.get_wavelengths()[:lim])

        if rad:
            if verbose: print("Converting to radiance... ", end="", flush="True")

            # convert from int to float
            image.data = image.data.astype(np.float32)

            # flag infs
            image.data[image.data == 65535.] = np.nan

            # apply dark reference
            if cls.dark is None:
                print("Warning: dark calibration not found; no dark correction was applied! Something smells dodgy...")
            else:
                dref = np.nanmean(cls.dark.data, axis=1)  # calculate dark reference
                image.data[:, :, :] -= dref[:, None, :lim]  # apply dark calibration

            # apply white reference (if specified)
            if not cls.white is None:
                # calculate white reference radiance
                white = cls.white.data.astype(np.float32) - dref[:, None, :]
                refl = np.ones(white.shape[-1])  # assume pure white

                # apply white reference
                cfac = refl[None, :] / np.nanmean( white, axis=1 )
                image.data[:, :, :] *= cfac[:, None, :lim]
                white = white[:,:,:lim] * cfac[:, None, :lim]


            if verbose: print("DONE.")

        ####################################################################
        # replace bad (nan) pixels with an average of the surrounding ones
        ####################################################################
        if bpr:
            if verbose: print("Filtering bad pixels... ", end="", flush="True")
            invalids = np.argwhere(np.isnan(image.data) | np.isinf(image.data))  # search for bad pixels
            for px, py, band in invalids:
                n = 0
                sum = 0
                for xx in range(px - 1, px + 2):
                    for yy in range(py - 1, py + 2):
                        if xx == px and yy == py: continue  # skip invalid pixel
                        if xx < 0 or yy < 0 or xx >= image.data.shape[0] or yy >= image.data.shape[
                            1]: continue  # skip out of bounds pixels
                        if image.data[xx][yy][band] == np.nan or image.data[xx][yy][
                            band] == np.inf: continue  # maybe neighbour is nan also
                        n += 1
                        sum += image.data[xx][yy][band]
                if n > 0: sum /= n  # do averaging
                image.data[px][py][band] = sum
            if verbose: print("DONE.")

        # Denoise LWIR along sensor plane
        image.data = median_filter(image.data, size=(3, 1, 3), mode="mirror")
        white = median_filter(white, size=(3,1,3), mode='mirror') # also apply to white panel for noise estimation

        # also estimate noise per-band (useful for eg., MNFs)
        noise = np.nanstd(white, axis=(0, 1))
        image.header['band noise'] = noise

        # rotate image so that scanning direction is horizontal rather than vertical)
        image.data = np.rot90(image.data)  # np.transpose(remap, (1, 0, 2))
        image.data = np.flip(image.data, axis=1)
        image.set_band_names(None)  # delete band names as they get super annoying

    @classmethod
    def correct_folder(cls, path, **kwds):

        """
        Many sensors use simple/common data structures to store data/headers/dark reference etc. Hence it is often easiest
        to pass an output folder to the sensor for correction.

        Args:
            path (str): a path to the folder containing the sensor specific data.
            **kwds: keywords are passed directly to correct_image, except for:

                - verbose = True if print outputs should be made to update progress. Default is True.

        Returns:
            A hyImage to which all sensor-specific corrections have been applied. Note that this will generally not include
           topographic or atmospheric corrections.

        """

        verbose = kwds.get("verbose", True)
        kwds["verbose"] = verbose

        imgs = [str(p) for p in Path(path).rglob("capture/*.hdr")]  # all image files [including data]
        dark = [str(p) for p in Path(path).rglob("capture/DARKREF*.hdr")]  # dark reference file
        white = [str(p) for p in Path(path).rglob("capture/WHITEREF*.hdr")]  # an white reference data (core scanner)
        refl = [str(p) for p in Path(path).rglob("capture/REFL*.hdr")]  # any processed reflectance data (SiSu Rock)
        for d in dark:
            del imgs[imgs.index(d)]
        for w in white:
            del imgs[imgs.index(w)]
        for r in refl:
            del imgs[imgs.index(r)]

        if len(imgs) > 1 or len(
            dark) > 1: assert False, "Error - multiple scenes found in folder. Double check file path..."
        if len(imgs) == 0 or len(
            dark) == 0: assert False, "Error - no image or dark calibration found in folder. Double check file path... %s" % path

        if verbose: print('\nLoading image %s' % imgs[0])

        # load image
        image = io.load(imgs[0])
        OWL.set_dark_ref(dark[0])

        if len(white) > 0:  # white reference exists
            OWL.set_white_ref(white[0])

        # correct
        OWL.correct_image(image, **kwds)

        # return corrected image
        return image



from gfit import mgauss
import numpy as np
import hylite
import skimage.data as skdata
from skimage.transform import rotate, swirl, rescale


def randomSpectra(wav: np.ndarray,
                  f: list = [2245., 2340.],
                  d: list = [0.2, 0.3],
                  w: list = [50, 100],
                  a: list = 0.9,
                  nrand: list = 15,
                  noise: list = 0.05,
                  slope: list = 0.2):
    """
    Generate a random spectra.
    :param wav: The wavelengths to generate a spectra across.
    :param f: Fixed absorption feature positions.
    :param d: Fixed absorption feature depths.
    :param w: Fixed absorption feature widths.
    :param a: Overall material brightness (from 0 to 1). Default is 0.9.
    :param nrand: Number of random features to add.
    :param noise: Depth of random features and additional white noise.
    :param slope: Slope of overall spectra.
    :return: A synthetic spectra
    """
    # generate initial (pure) spectra
    R = (1 - slope * (wav - wav[0]) / (wav[-1] - wav[0]))  # slope
    R2 = np.zeros_like(wav)
    mgauss(wav, R2, d, f, w)
    R = R - R2

    if nrand > 0:
        mgauss(wav, R2, [np.random.uniform(0., noise) for i in range(nrand)],
               [np.random.uniform(wav[0], wav[-1]) for i in range(nrand)],
               [np.random.uniform(10 * (wav[1] - wav[0]), (wav[-1] - wav[0])) for i in range(nrand)])
        R = a * (R - R2)
    return np.clip(R + np.random.rand(len(wav)) * noise, 0., 1.)


def genImage(wav: np.ndarray = np.linspace(2100., 2400., 200),
             A: dict = dict(f=[2200., 2340.], a=1.0, nrand=0),
             B: dict = dict(f=[2160., 2210.], a=0.5, nrand=0),
             C: dict = dict(f=[2250., 2350.], a=0.2, nrand=0),
             flip: bool = False, seed: int = 42, noise: float = 0.05):
    """
    Use the Bricks() and Stones() images in skimage.data to create a synthetic hyperspectral image. Useful
    for testing or demonstration purposes.
    :param wav: Numpy array containing wavelengths of the desired HSI image. Default is np.linspace(2100., 2400., 200).
    :param A: A dictionary containing arguments passed to randomSpectra(...) for phase A. Default is {f=[2200., 2340.], a=1.0}.
    :param B: A dictionary containing arguments passed to randomSpectra(...) for phase B. Default is {f=[2160., 2200.], a=0.5}.
    :param C: A dictionary containing arguments passed to randomSpectra(...) for phase C. Default is {f=[2250., 2350.], a=0.2}.
    :param flip: True if the order of latent variables should be reversed. Useful for generating different styles of image.
    :param seed: An integer seed for numpy.random(...). Default is 42.
    :param noise: Amount of white noise to add. Default is 0.05.
    :return: The synthetic hyperspectral image
    """

    np.random.seed(seed)

    # Define two latent variables
    lA = (255 - np.clip(2 * skdata.brick(), 0, 255)) / 255.
    lA = swirl(rotate(lA, 15, mode='symmetric'), strength=2, radius=1024)
    lB = swirl(rescale(np.tile(skdata.grass() / 255., (2, 2)), 0.5), strength=2, radius=1024)

    # Combine into three closed abundances (sum to 1)
    ab = np.dstack([lA, lB, 1.0 - (lA + lB)])
    ab[ab < 0] = 0.01
    ab = ab ** 2  # enhance contrast a little
    ab /= np.sum(ab, axis=-1)[..., None] # ensure abundances sum to 1
    ab = ab[:-2, :, :] # adjust abundance so that it isn't square (test data should never be square!)
    if flip:
        ab = ab[:,:,::-1]

    # Sample endmember spectra
    eA = randomSpectra(wav, **A)  # Phase A
    eB = randomSpectra(wav, **B)  # Phase B
    eC = randomSpectra(wav, **C)  # Phase C

    # mix to generate fake HSI data
    hsi = (ab.reshape(-1, 3) @ np.array([eA, eB, eC])).reshape((ab.shape[:-1]) + (len(wav),))
    hsi += np.random.rand(*hsi.shape) * noise
    # put image and abundances in a HyImage object and return
    hsi = hylite.HyImage(hsi)
    hsi.set_wavelengths(wav)
    A = hylite.HyImage(ab)
    A.set_band_names(['mineral_A', 'mineral_B', 'mineral_C'])
    return hsi, A

